# pylint: disable=wildcard-import
"""Video action recognition."""
from __future__ import absolute_import
from .actionrec_vgg16 import *
from .actionrec_inceptionv1 import *
from .actionrec_inceptionv3 import *
from .actionrec_resnetv1b import *
from .c3d import *
from .p3d import *
from .r2plus1d import *
from .i3d_resnet import *
from .i3d_inceptionv1 import *
from .i3d_inceptionv3 import *
from .slowfast import *
