"""Presets of sophisticated data transforms used in various papers."""
from . import ssd
from . import rcnn
from . import yolo
from . import center_net
from . import imagenet
from . import simple_pose
from . import segmentation
